/***
 * Fractal ADL Parser
 * Copyright (C) 2002-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.deployment.scheduling.core.lib.BasicScheduler;
import org.objectweb.fractal.adl.arguments.ArgumentComponentLoader;
import org.objectweb.fractal.adl.arguments.ArgumentLoader;
import org.objectweb.fractal.adl.attributes.AttributeCompiler;
import org.objectweb.fractal.adl.attributes.AttributeLoader;
import org.objectweb.fractal.adl.attributes.JavaAttributeBuilder;
import org.objectweb.fractal.adl.bindings.BindingCompiler;
import org.objectweb.fractal.adl.bindings.JavaBindingBuilder;
import org.objectweb.fractal.adl.bindings.TypeBindingLoader;
import org.objectweb.fractal.adl.components.ComponentCompiler;
import org.objectweb.fractal.adl.components.JavaComponentBuilder;
import org.objectweb.fractal.adl.components.PrimitiveComponentCompiler;
import org.objectweb.fractal.adl.implementations.ImplementationCompiler;
import org.objectweb.fractal.adl.implementations.ImplementationLoader;
import org.objectweb.fractal.adl.implementations.JavaImplementationBuilder;
import org.objectweb.fractal.adl.interfaces.InterfaceLoader;
import org.objectweb.fractal.adl.types.JavaTypeBuilder;
import org.objectweb.fractal.adl.types.TypeCompiler;
import org.objectweb.fractal.adl.types.TypeLoader;
import org.objectweb.fractal.adl.xml.XMLLoader;

/**
 * Provides static methods to get a {@link Factory} component.
 */

public class FactoryFactory {
  
  public final static String FRACTAL_BACKEND =
    "org.objectweb.fractal.adl.FractalBackend";
  
  public final static String STATIC_FRACTAL_BACKEND =
    "org.objectweb.fractal.adl.StaticFractalBackend";
  
  public final static String JAVA_BACKEND =
    "org.objectweb.fractal.adl.JavaBackend";
  
  public final static String STATIC_JAVA_BACKEND =
    "org.objectweb.fractal.adl.StaticJavaBackend";
  
  private static Factory FACTORY;
  
  /**
   * Returns a bootstrap {@link Factory}, with a Java backend.
   * 
   * @return a bootstrap factory. 
   */
  
  public static Factory getFactory () {
    if (FACTORY != null) {
      return FACTORY;
    }
    
    BasicFactory r = new BasicFactory();
        
    XMLLoader xmll = new XMLLoader();
    ArgumentLoader argl = new ArgumentLoader();
    InterfaceLoader itfl = new InterfaceLoader();
    TypeLoader typl = new TypeLoader();
    ImplementationLoader impll = new ImplementationLoader();
    AttributeLoader attrl1 = new AttributeLoader();
    ArgumentComponentLoader compl = new ArgumentComponentLoader();
    TypeBindingLoader bindl = new TypeBindingLoader();
    // necessary to check inherited/overriden attributes
    AttributeLoader attrl2 = new AttributeLoader();
    
    TypeCompiler typc = new TypeCompiler();
    ImplementationCompiler implc = new ImplementationCompiler();
    PrimitiveComponentCompiler compc = new PrimitiveComponentCompiler();
    BindingCompiler bindc = new BindingCompiler();
    AttributeCompiler attrc = new AttributeCompiler();
    ComponentCompiler allc = new ComponentCompiler();
        
    JavaTypeBuilder typb = new JavaTypeBuilder();
    JavaImplementationBuilder implb = new JavaImplementationBuilder();
    JavaComponentBuilder compb = new JavaComponentBuilder();
    JavaBindingBuilder bindb = new JavaBindingBuilder();
    JavaAttributeBuilder attrb = new JavaAttributeBuilder();

    BasicScheduler s = new BasicScheduler();
    
    typc.bindFc(TypeCompiler.BUILDER_BINDING, typb);
    implc.bindFc(ImplementationCompiler.BUILDER_BINDING, implb);
    compc.bindFc(PrimitiveComponentCompiler.BUILDER_BINDING, compb);
    bindc.bindFc(BindingCompiler.BUILDER_BINDING, bindb);
    attrc.bindFc(AttributeCompiler.BUILDER_BINDING, attrb);
    
    argl.clientLoader = xmll;
    itfl.clientLoader = argl;
    typl.clientLoader = itfl;
    impll.clientLoader = typl;
    attrl1.clientLoader = impll;
    compl.clientLoader = attrl1;
    bindl.clientLoader = compl;
    attrl2.clientLoader = bindl;
    
    allc.bindFc(ComponentCompiler.PRIMITIVE_COMPILERS_BINDING+"0", typc);
    allc.bindFc(ComponentCompiler.PRIMITIVE_COMPILERS_BINDING+"1", implc);
    allc.bindFc(ComponentCompiler.PRIMITIVE_COMPILERS_BINDING+"2", compc);
    allc.bindFc(ComponentCompiler.PRIMITIVE_COMPILERS_BINDING+"3", bindc);
    allc.bindFc(ComponentCompiler.PRIMITIVE_COMPILERS_BINDING+"4", attrc);
    
    r.bindFc(BasicFactory.LOADER_BINDING, attrl2);
    r.bindFc(BasicFactory.COMPILER_BINDING, allc);
    r.bindFc(BasicFactory.SCHEDULER_BINDING, s);
    
    FACTORY = r;
    return r;
  }
  
  /**
   * Returns a {@link Factory} with the given backend.
   * 
   * @param backend the desired backend.
   * @return a {@link Factory} with the given backend.
   * @throws ADLException if the factory cannot be created.
   */
  
  public static Factory getFactory (final String backend) throws ADLException {
    return getFactory(backend, new HashMap());
  }

  /**
   * Returns a {@link Factory} with the given backend.
   * 
   * @param backend the desired backend.
   * @param context additional optional information.
   * @return a {@link Factory} with the given backend.
   * @throws ADLException if the factory cannot be created.
   */
  
  public static Factory getFactory (final String backend, final Map context) 
    throws ADLException 
  {
    return getFactory("org.objectweb.fractal.adl.BasicFactory", backend, context);
  }
  
  /**
   * Returns a {@link Factory} with the given implementation and backend.
   * 
   * @param factory the name of the desired factory.
   * @param backend the desired backend.
   * @param context additional optional information.
   * @return a {@link Factory} with the given backend.
   * @throws ADLException if the factory cannot be created.
   */
  
  public static Factory getFactory (
    final String factory, 
    final String backend, 
    final Map context) throws ADLException
  {
    context.put("backend", backend);
    Factory f = getFactory();
    Map c = (Map)f.newComponent(factory, context);
    return (Factory)c.get("factory");
  }
}
