/***
 * Fractal ADL Parser
 * Copyright (C) 2002-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl.components;

import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Type;
import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.AbstractNode;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.AbstractLoader;
import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.NodeClassLoader;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Iterator;
import java.util.HashMap;
import java.util.Set;

/**
 * A {@link org.objectweb.fractal.adl.Loader} to check {@link Component} nodes 
 * in definitions. This loader checks sub component names, loads the definitons
 * referenced by the component nodes (and the "extends" attribute), and merges
 * all these definitions into a single one (by following inheritance rules).
 */

public class ComponentLoader 
  extends AbstractLoader implements ComponentLoaderAttributes 
{

  /**
   * The names of the "name" attribute for each AST node type. This map 
   * associates the names of the "name" attribute, used to detect overriden 
   * elements, to each AST node type that has such an attribute.
   */

  private Map NAME_ATTRIBUTES;

  /**
   * The class loader used to merge AST node classes. This class merging is
   * necessary when merging several ASTs with different sources (eg different
   * DTDs).
   */
  
  private MergeClassLoader mergeClassLoader;
  
  // -------------------------------------------------------------------------
  // Constructor
  // -------------------------------------------------------------------------

  public ComponentLoader () {
    NAME_ATTRIBUTES = new HashMap();
    NAME_ATTRIBUTES.put("component", "name");
    NAME_ATTRIBUTES.put("interface", "name");
    NAME_ATTRIBUTES.put("binding", "from");
    NAME_ATTRIBUTES.put("attribute", "name");
    NAME_ATTRIBUTES.put("coordinates", "name");
    mergeClassLoader = new MergeClassLoader(getClass().getClassLoader());
  }

  // -------------------------------------------------------------------------
  // Implementation of the ComponentLoaderAttributes interface
  // -------------------------------------------------------------------------

  public String getNameAttributes () {
    StringBuffer b = new StringBuffer();
    Iterator i = NAME_ATTRIBUTES.entrySet().iterator();
    while (i.hasNext()) {
      Map.Entry e = (Map.Entry)i.next();
      b.append((String)e.getKey());
      b.append(' ');
      b.append((String)e.getValue());
      b.append(' ');
    }
    return b.toString();
  }
  
  public void setNameAttributes (String nameAttributes) {
    NAME_ATTRIBUTES.clear();
    String key = null;
    int p = nameAttributes.indexOf(' ');
    while (p != -1) {
      String s = nameAttributes.substring(0, p);
      if (key == null) {
        key = s;
      } else {
        NAME_ATTRIBUTES.put(key, s);
        key = null;
      }
      nameAttributes = nameAttributes.substring(p + 1);
      p = nameAttributes.indexOf(' ');
    }
  }
  
  // -------------------------------------------------------------------------
  // Implementation of the Loader interface
  // -------------------------------------------------------------------------

  public Definition load (final String name, final Map context) 
    throws ADLException 
  {
    return load(new ArrayList(), name, context);
  }
  
  public Definition load (
    final List loaded,
    final String name,
    final Map context) throws ADLException 
  {
    if (loaded.contains(name)) {
      throw new ADLException("Cycle in definition references: " + loaded, null);
    }
    List l = new ArrayList(loaded);
    l.add(name);
    Definition d = clientLoader.load(name, context);
    if (d instanceof ComponentDefinition) {
      ComponentDefinition container = (ComponentDefinition)d;
      normalizeComponentContainer(container);
      if (container.getExtends() != null) {
        try {
          d = (Definition)merge((Node)d, (Node)resolveDefinitions(l, container.getExtends(), context));
        } catch (ADLException e) {
          throw new ADLException(
            "Cannot load super definition(s)", (Node)container, e);
        }
        container = (ComponentDefinition)d;
        container.setExtends(null);
      }
      resolveComponentContainer(l, container, container, context);
    }
    return d;
  }

  // -------------------------------------------------------------------------
  // First pass: normalization
  // -------------------------------------------------------------------------

  public void normalizeComponentContainer (final ComponentContainer container) 
    throws ADLException
  {
    Set names = new HashSet();
    Component[] comps = container.getComponents();
    for (int i = 0; i < comps.length; i++) {
      Component comp = comps[i];
      String name = comp.getName();
      if (name == null) {
        throw new ADLException("Component name missing", (Node)comp);
      }
      if (names.contains(name)) {
        throw new ADLException(
          "Duplicated component name '" + name + "'", (Node)comp);
      } else {
        names.add(name);
      }
      normalizeComponentContainer(comp);
      /*if (name.indexOf('/') != -1) {
        container.removeComponent(comp);
        addComponent(container, comp);
      }*/
    }
  }
  
  // this feature has been removed
  
  /*public void addComponent (
    final ComponentContainer container, Component comp) throws ADLException
  {
    String name = comp.getName();
    int p = name.indexOf('/');
    if (p == -1) {
      Component c = getComponent(container, name);
      if (c != null) {
        replaceComponent(container, c, (Component)merge((Node)comp, (Node)c));
      } else {
        container.addComponent(comp);
      }
    } else {
      Component parent = getComponent(container, name.substring(0, p));
      if (parent == null) {
        try {
          parent = (Component)comp.getClass().newInstance();
        } catch (InstantiationException e) {
          throw new Error("Internal error");
        } catch (IllegalAccessException e) {
          throw new Error("Internal error");
        }
        parent.setName(name.substring(0, p));
        container.addComponent(parent);
      }
      comp.setName(name.substring(p + 1));
      addComponent(parent, comp);
    }
  }*/

  // -------------------------------------------------------------------------
  // Second pass: definition references resolution
  // -------------------------------------------------------------------------

  public void resolveComponentContainer (
    final List loaded,
    final ComponentContainer topLevelDefinition,
    final ComponentContainer container,
    final Map context) throws ADLException
  {
    Component[] comps = container.getComponents();
    for (int i = 0; i < comps.length; i++) {
      Component comp = comps[i];
      resolveComponentContainer(loaded, topLevelDefinition, comp, context);
      String definition = comp.getDefinition();
      comp.setDefinition(null);
      ((Node)comp).astSetDecoration("definition", definition);
      if (definition != null) {
        if (definition.indexOf('/') != -1) {
          // shared component
          if (definition.startsWith("./")) {
            definition = definition.substring(2);
          }
          Component c = getPathComponent(topLevelDefinition, definition);
          if (c == null) {
            throw new ADLException(
              "No such component", (Node)comp);
          }
          if (!c.getName().equals(comps[i].getName())) {
            throw new ADLException(
              "Shared components with distinct names not yet supported", (Node)comp);
          }
          Map replacements = new HashMap();
          replacements.put(comp, c);
          replaceComponents(topLevelDefinition, replacements);
        } else {
          Definition d;
          try {
            d = resolveDefinitions(loaded, definition, context);
          } catch (ADLException e) {
            throw new ADLException(
              "Cannot load referenced definition(s)", (Node)comp, e);
          }
          Map replacements = new HashMap();
          merge((Node)comp, (Node)d, replacements);
          replaceComponents(topLevelDefinition, replacements);
        }
      }
    }
  }

  public Definition resolveDefinitions (
    final List loaded,
    final String nameList, 
    final Map context) throws ADLException
  {
    int p = nameList.lastIndexOf(',');
    if (p == -1) {
      return load(loaded, nameList, context);
    } else {
      Definition d = resolveDefinitions(loaded, nameList.substring(0, p), context);
      Definition e = load(loaded, nameList.substring(p + 1), context);
      return (Definition)merge((Node)e, (Node)d);
    }
  }

  public Component getComponent (
    final ComponentContainer container,
    final String name)
  {
    Component[] comps = container.getComponents();
    for (int i = 0; i < comps.length; i++) {
      Component comp = comps[i];
      if (comp.getName().equals(name)) {
        return comp;
      }
    }
    return null;
  }

  public Component getPathComponent (
    final ComponentContainer container,
    final String name)
  {
    int p = name.indexOf('/');
    if (p == -1) {
      return getComponent(container, name);
    } else {
      Component parent = getComponent(container, name.substring(0, p));
      return getPathComponent(parent, name.substring(p + 1));
    }
  }

  public ComponentContainer replaceComponents (
    final ComponentContainer container,
    final Map replacements)
  {
    if (replacements.get(container) != null) {
      return (ComponentContainer)replacements.get(container);
    }
    Component[] comps = container.getComponents();
    for (int i = 0; i < comps.length; i++) {
      container.removeComponent(comps[i]);
    }
    for (int i = 0; i < comps.length; i++) {
      container.addComponent(
        (Component)replaceComponents(comps[i], replacements));
    }
    return container;
  }

  // -------------------------------------------------------------------------
  // Utility methods: inheritance resolution
  // -------------------------------------------------------------------------

  public Node merge (final Node elem, final Node superElem) throws ADLException {
    return merge(elem, superElem, null);
  }
      
  public Node merge (
    final Node elem, 
    final Node superElem,
    final Map replacements) throws ADLException
  {
    Map infos = new HashMap();
    computeMergeInfos(elem, superElem, infos);
    createMergedNodes(infos);
    Node n = initMergedNodes((MergeInfo)infos.get(elem), infos);
    if (replacements != null) {
      Iterator i = infos.entrySet().iterator();
      while (i.hasNext()) {
        Map.Entry e = (Map.Entry)i.next();
        replacements.put(e.getKey(), ((MergeInfo)e.getValue()).result);
      }
    }
    return n;
  }

  private void computeMergeInfos (
    final Node elem, 
    final Node superElem, 
    final Map infos) throws ADLException
  {
    MergeInfo info = (MergeInfo)infos.get(elem);
    if (info == null) {
      info = new MergeInfo();
      info.nodes.add(elem);
      infos.put(elem, info);
    }
    if (info.nodes.contains(superElem)) {
      return;
    } else {
      info.nodes.add(superElem);
      infos.put(superElem, info);
    }
    
    Set nodeTypes = new HashSet();
    nodeTypes.addAll(Arrays.asList(elem.astGetNodeTypes()));
    nodeTypes.addAll(Arrays.asList(superElem.astGetNodeTypes()));
    
    Iterator i = nodeTypes.iterator();
    while (i.hasNext()) {
      String nodeType = (String)i.next();
      Node[] superNodes = superElem.astGetNodes(nodeType);      
      Node[] nodes = elem.astGetNodes(nodeType);
      if (superNodes == null) {
        superNodes = new Node[0];
      }
      if (nodes == null) {
        nodes = new Node[0];
      }
      
      String nameAttr = (String)NAME_ATTRIBUTES.get(nodeType);

      if (nameAttr == null) {
        if (superNodes.length > 0 && superNodes[0] != null) {
          if (nodes.length == 0 || nodes[0] == null) {
            computeMergeInfos(superNodes[0], infos);
          } else {
            computeMergeInfos(nodes[0], superNodes[0], infos);
          }
        } else {
          if (nodes.length > 0 && nodes[0] != null) {
            computeMergeInfos(nodes[0], infos);
          }
        }
      } else {
        for (int k = 0; k < superNodes.length; ++k) {
          String superName = (String)superNodes[k].astGetAttributes().get(nameAttr);
          int index = -1;
          for (int l = 0; l < nodes.length; ++l) {
            String name = (String)nodes[l].astGetAttributes().get(nameAttr);
            if (name.equals(superName)) {
              index = l;
              break;
            }
          }
          if (index == -1) {
            computeMergeInfos(superNodes[k], infos);
          } else {
            computeMergeInfos(nodes[index], superNodes[k], infos);
          }
        }
        
        for (int l = 0; l < nodes.length; ++l) {
          String name = (String)nodes[l].astGetAttributes().get(nameAttr);
          int index = -1;
          for (int k = 0; k < superNodes.length; ++k) {
            String superName = (String)superNodes[k].astGetAttributes().get(nameAttr);
            if (superName.equals(name)) {
              index = k;
              break;
            }
          }
          if (index == -1) {
            computeMergeInfos(nodes[l], infos);
          }
        }
      }
    }
  }
  
  private void computeMergeInfos (final Node node, final Map infos)
    throws ADLException 
  {
    MergeInfo info = (MergeInfo)infos.get(node);
    if (info != null) {
      return;
    }
    info = new MergeInfo();
    info.nodes.add(node);
    infos.put(node, info);
    String[] types = node.astGetNodeTypes();
    for (int i = 0; i < types.length; ++i) {
      String type = types[i];
      Node[] subNodes = node.astGetNodes(type);
      for (int j = 0; j < subNodes.length; j++) {
        Node subNode = subNodes[j];
        if (subNode != null) {
          computeMergeInfos(subNode, infos);
        }
      }
    }
  }
  
  private void createMergedNodes (final Map infos) 
    throws ADLException 
  {
    Iterator i = infos.values().iterator();
    while (i.hasNext()) {
      MergeInfo info = (MergeInfo)i.next();
      if (info.result != null) {
        continue;
      }
      
      Node elem = (Node)info.nodes.get(0);
      List classes = new ArrayList();
      classes.add(elem.getClass());
      
      boolean ok = true;
      for (int j = 1; j < info.nodes.size(); ++j) {
        Class sec = info.nodes.get(j).getClass();
        if (sec.getClassLoader() != ((Class)classes.get(0)).getClassLoader()) {
          ok = false;
        }
        classes.add(sec);
      }
      
      if (ok) {
        try {
          info.result = (Node)elem.getClass().newInstance();
        } catch (Exception e) {
          throw new ADLException("Cannot merge ASTs", elem, e);
        }
      } else {
        try {
          int hash = elem.getClass().hashCode();
          for (int j = 1; j < classes.size(); ++j) {
            hash = 17*(hash + classes.get(j).hashCode());
          }
          String code = Integer.toHexString(hash);
          String name = "org.objectweb.fractal.adl.merged.Merged" + code;
          Class merged = mergeClassLoader.mergeClass(
            name,
            elem.astGetType(), 
            (Class[])classes.toArray(new Class[classes.size()]));
          info.result = (Node)merged.newInstance();
        } catch (Exception e) {
          throw new ADLException("Cannot merge AST classes", elem, e);
        }
      }
    }    
  }

  private Node initMergedNodes (final MergeInfo info, final Map infos) {
    if (info.done) {
      return info.result;
    }
    info.done = true;

    Node elem = (Node)info.nodes.get(0);
    Node result = info.result;
    
    // merges source
    result.astSetSource(elem.astGetSource());

    // merges attributes
    Map resultAttrs = elem.astGetAttributes();
    for (int i = 1; i < info.nodes.size(); ++i) {
      Map superAttrs = ((Node)info.nodes.get(i)).astGetAttributes();
      Iterator j = superAttrs.keySet().iterator();
      while (j.hasNext()) {
        String name = (String)j.next();
        String superValue = (String)superAttrs.get(name);
        String value = (String)resultAttrs.get(name);
        if (superValue != null && value == null) {
          resultAttrs.put(name, superValue);
        }
      }
    }
    result.astSetAttributes(resultAttrs);
    
    // merges decorations
    Map resultDecors = elem.astGetDecorations();
    for (int i = 1; i < info.nodes.size(); ++i) {
      Map superDecors = ((Node)info.nodes.get(i)).astGetDecorations();
      resultDecors.putAll(superDecors);
    }
    result.astSetDecorations(resultDecors);
    
    // merges sub nodes    
    Set nodeTypes = new HashSet();
    nodeTypes.addAll(Arrays.asList(elem.astGetNodeTypes()));
    for (int i = 1; i < info.nodes.size(); ++i) {
      nodeTypes.addAll(
        Arrays.asList(((Node)info.nodes.get(i)).astGetNodeTypes()));
    }
    
    Iterator i = nodeTypes.iterator();
    while (i.hasNext()) {
      String nodeType = (String)i.next();
      String nameAttr = (String)NAME_ATTRIBUTES.get(nodeType);
      
      List resultNodes = new ArrayList();
      Node[] nodes = elem.astGetNodes(nodeType);
      if (nodes == null) {
        nodes = new Node[0];
      }
      for (int j = 0; j < nodes.length; ++j) {
        if (nodes[j] != null) {
          MergeInfo inf = (MergeInfo)infos.get(nodes[j]);
          resultNodes.add(initMergedNodes(inf, infos));
        }
      }
      
      for (int j = 1; j < info.nodes.size(); ++j) {
        Node superElem = (Node)info.nodes.get(j);
        Node[] superNodes = superElem.astGetNodes(nodeType);
        if (superNodes == null) {
          continue;
        }
        
        if (nameAttr == null) {
          if (superNodes.length > 0 && superNodes[0] != null) {
            if (nodes.length == 0 || nodes[0] == null) {
              MergeInfo inf = (MergeInfo)infos.get(superNodes[0]);
              resultNodes.add(initMergedNodes(inf, infos));
            } else {
              MergeInfo inf = (MergeInfo)infos.get(nodes[0]);
              resultNodes.set(0, initMergedNodes(inf, infos));
            }
          }
        } else {
          for (int k = 0; k < superNodes.length; ++k) {
            String superName = (String)superNodes[k].astGetAttributes().get(nameAttr);
            int index = -1;
            for (int l = 0; l < nodes.length; ++l) {
              String name = (String)nodes[l].astGetAttributes().get(nameAttr);
              if (name.equals(superName)) {
                index = l;
                break;
              }
            }
            if (index == -1) {
              MergeInfo inf = (MergeInfo)infos.get(superNodes[k]);
              resultNodes.add(initMergedNodes(inf, infos));
            } else {
              MergeInfo inf = (MergeInfo)infos.get(nodes[index]);
              resultNodes.set(index, initMergedNodes(inf, infos));
            }
          }
        }
      }
      
      for (int j = 0; j < resultNodes.size(); ++j) {
        result.astAddNode((Node)resultNodes.get(j));
      }
    }
    
    return info.result;
  }

  private static class MergeInfo {    
        
    List nodes = new ArrayList();
    
    Node result;
    
    boolean done;
  }
  
  private static class MergeClassLoader extends NodeClassLoader {
            
    public MergeClassLoader (final ClassLoader parent) {
      super(parent);
    }
    
    public Class mergeClass (final String name, final String astNodeName, final Class[] classes)
      throws ClassNotFoundException 
    {
      try {
        return loadClass(name);
      } catch (ClassNotFoundException e) {        
      }
            
      Set itfs = new HashSet();
      for (int i = 0; i < classes.length; ++i) {
        Class[] citfs = classes[i].getInterfaces();
        for (int j = 0; j < citfs.length; ++j) {
          itfs.add(Type.getInternalName(citfs[j]));
        }
      }
      
      ClassWriter cw = generateClass(
        name, 
        astNodeName,
        Type.getInternalName(AbstractNode.class), 
        (String[])itfs.toArray(new String[itfs.size()]));
      
      /* DEBUG
      try {
        java.io.FileOutputStream fos =
          new java.io.FileOutputStream(name + ".class");
        fos.write(cw.toByteArray());
        fos.close();
      } catch (Exception _) {
      }
      */
      return defineClass(name, cw.toByteArray());
    }
  }
}
