/***
 * Fractal ADL Parser
 * Copyright (C) 2002-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl.types;

import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.AbstractLoader;
import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.components.ComponentDefinition;
import org.objectweb.fractal.adl.interfaces.Interface;
import org.objectweb.fractal.adl.interfaces.InterfaceContainer;

/**
 * A {@link org.objectweb.fractal.adl.Loader} to check {@link TypeInterface} 
 * nodes in definitions. This loader checks that the Java interfaces specified
 * in these nodes exist.
 */

public class TypeLoader extends AbstractLoader {
  
  // --------------------------------------------------------------------------
  // Implementation of the Loader interface
  // --------------------------------------------------------------------------
  
  public Definition load (final String name, final Map context) 
    throws ADLException 
  {
    Definition d = clientLoader.load(name, context);
    boolean extend = false;
    if (d instanceof ComponentDefinition) {
      extend = ((ComponentDefinition)d).getExtends() != null;
    }
    checkNode(d, extend);
    return d;
  }

  // --------------------------------------------------------------------------
  // Checking methods
  // --------------------------------------------------------------------------
  
  private void checkNode (final Object node, final boolean extend) 
    throws ADLException 
  {
    if (node instanceof InterfaceContainer) {
      checkInterfaceContainer((InterfaceContainer)node, extend);
    }
    if (node instanceof ComponentContainer) {
      Component[] comps = ((ComponentContainer)node).getComponents();
      for (int i = 0; i < comps.length; i++) {
        checkNode(comps[i], extend);
      }
    }
  }

  private void checkInterfaceContainer (
    final InterfaceContainer container,
    final boolean extend) throws ADLException
  {
    ClassLoader cl = getClass().getClassLoader();
    Interface[] itfs = container.getInterfaces();
    for (int i = 0; i < itfs.length; i++) {
      Interface itf = itfs[i];
      if (itf instanceof TypeInterface) {
        String signature = ((TypeInterface)itf).getSignature();
        if (signature == null) {
          if (!extend) {
            throw new ADLException("Signature missing", (Node)itf);
          }
        } else {
          try {
            cl.loadClass(signature);
          } catch (ClassNotFoundException e) {
            throw new ADLException(
              "Invalid signature '" + signature + "'", (Node)itf, e);
          }
        }
        String role = ((TypeInterface)itf).getRole();
        if (role == null) {
          if (!extend) {
            throw new ADLException("Role missing", (Node)itf);
          }
        } else {
          if (!role.equals("client") && !role.equals("server")) {
            throw new ADLException("Invalid role '" + role + "'", (Node)itf);
          }  
        }
        String contingency = ((TypeInterface)itf).getContingency();
        if (contingency != null) {
          if (!contingency.equals("mandatory") && !contingency.equals("optional")) {
            throw new ADLException("Invalid contingency '" + contingency + "'", (Node)itf);
          }
        }
        String cardinality = ((TypeInterface)itf).getCardinality();
        if (cardinality != null) {
          if (!cardinality.equals("singleton") && !cardinality.equals("collection")) {
            throw new ADLException("Invalid cardinality '" + cardinality + "'", (Node)itf);
          }
        }
      }
    }
  }
}
