/***
 * Fractal ADL Parser
 * Copyright (C) 2002-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl.xml;

import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.Parser;
import org.objectweb.fractal.adl.ParserException;
import org.objectweb.fractal.adl.xml.XMLNodeClassLoader;
import org.objectweb.fractal.adl.xml.DTDHandler;
import org.objectweb.fractal.adl.xml.XMLNode;

import java.io.InputStream;
import java.io.IOException;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.net.URL;
import java.net.MalformedURLException;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.SAXException;
import org.xml.sax.InputSource;
import org.xml.sax.Attributes;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.Locator;
import org.xml.sax.SAXParseException;

/**
 * An extensible XML parser that constructs strongly typed abstract syntax trees.
 */

public class XMLParser extends DefaultHandler implements Parser {

  /**
   * The parser factory used to create new parsers.
   */
  
  SAXParserFactory spf;
  
  /**
   * Associates an ASTClassLoader to each DTD.
   * Keys are DTD names, values are ASTClassLoader objects.
   */

  Map loaders;

  /**
   * Name of the file that is currently being parsed.
   */

  String file;

  /**
   * Loader used for the elements that are currently being parsed.
   */

  XMLNodeClassLoader loader;

  /**
   * Stack of the elements that are currently being parsed.
   */

  List stack;

  /**
   * Top level element of the elements that are currently being parsed.
   */

  XMLNode result;

  /**
   * The locator used to keep track of line numbers.
   */
  
  Locator locator;

  int line;
  
  // --------------------------------------------------------------------------
  // Constructor
  // --------------------------------------------------------------------------

  public XMLParser () {
    this(true);
  }
  
  public XMLParser (boolean validate) {
    loaders = new HashMap();
    stack = new ArrayList();
    spf = SAXParserFactory.newInstance();
    spf.setValidating(validate);
  }

  // --------------------------------------------------------------------------
  // Implementation of the Parser interface
  // --------------------------------------------------------------------------

  public Node parse (final InputStream is, final String name) 
    throws ParserException 
  {
    file = name;
    try {
      SAXParser sp = spf.newSAXParser();
      sp.parse(new InputSource(is), this);
      return result;
    } catch (IOException e) {
      throw new ParserException("Parser error (" + file + ":" + line + ")", e);
    } catch (ParserConfigurationException e) {
      throw new ParserException("Parser error (" + file + ":" + line + ")", e);
    } catch (SAXException e) {
      throw new ParserException("Parser error (" + file + ":" + line + ")", e);
    }
  }

  // --------------------------------------------------------------------------
  // Overriden DefaultHandler methods
  // --------------------------------------------------------------------------

  public void setDocumentLocator (final Locator locator) {
    super.setDocumentLocator(locator);
    this.locator = locator;
  }

  public InputSource resolveEntity (
    final String publicId,
    final String systemId) throws SAXException
  {
    // finds the AST class loader for the "systemID" DTD
    loader = (XMLNodeClassLoader)loaders.get(systemId);
    if (loader == null) {
      loader = new XMLNodeClassLoader(getClass().getClassLoader());
      loaders.put(systemId, loader);

      try {
        InputStream is;
        // gets an InputStream to read the DTD
        if (systemId.startsWith("classpath://")) {
          is = getClass().getClassLoader().getResourceAsStream(
            systemId.substring("classpath://".length()));
        } else if (systemId.startsWith("file:")) {
          is = new URL(systemId).openStream();
        } else {
          throw new SAXNotRecognizedException(
            "Unrecognized system identifier: " + systemId);
        }
        // verifies the DTD and initializes the AST class loader
        new DTDHandler().checkDTD(is, loader);
      } catch (MalformedURLException e) {
        throw new SAXException("Cannot find the DTD", e);
      } catch (IOException e) {
        throw new SAXException("Cannot read the DTD", e);
      } catch (ClassNotFoundException e) {
        throw new SAXException("Cannot check the DTD", e);
      }
    }

    if (systemId.startsWith("classpath://")) {
      return new InputSource(getClass().getClassLoader().getResourceAsStream(
        systemId.substring("classpath://".length())));
    } else {
      return null;
    }
  }

  public void startElement (
    final String uri,
    final String localName,
    final String qualifiedName,
    final Attributes attributes) throws SAXException
  {
    String xmlNodeClassName = loader.getASTClassName(qualifiedName);
    XMLNode o;
    try {
      o = (XMLNode)loader.loadClass(xmlNodeClassName).newInstance();
    } catch (Exception e) {
      throw new SAXException("Internal error", e);
    }
    o.astSetSource(file + ":" + locator.getLineNumber());
    o.xmlSetAttributes(attributes);
    if (stack.size() == 0) {
      result = o;
    } else {
      Object p = stack.get(stack.size() - 1);
      ((XMLNode)p).xmlAddNode(qualifiedName, o);
    }
    stack.add(o);
  }

  public void endElement (
    final String uri,
    final String localName,
    final String qualifiedName) throws SAXException
  {
    stack.remove(stack.size() - 1);
  }

  public void error (final SAXParseException e) throws SAXException {
    line = locator.getLineNumber();
    throw e;
  }

  public void fatalError (final SAXParseException e) throws SAXException {
    line = locator.getLineNumber();
    throw e;
  }
}
